@@ -378,6 +378,18 @@ static pl_penalty_t __pl_position_get_pos_simple(
 		/* TA:  20090218:  Try and keep the window on-screen if we
 		 * can.
 		 */
+
+		/* TA:  20120316:  Imply the working-area when under the mouse -- this
+		 * brings it in-line with making the EWMH working area the default.
+		 * Note that "UnderMouse" is a special case, deliberately.  All other
+		 * PositionPlacement commands are deliberately NOT subject to ewmhiwa
+		 * options.
+		 */
+		EWMH_GetWorkAreaIntersection(
+			arg->place_fw, (int *)&arg->screen_g.x, (int *)&arg->screen_g.y,
+			(int *)&arg->screen_g.width,
+			(int *)&arg->screen_g.height, EWMH_USE_WORKING_AREA);
+
 		if (ret_p->x + arg->place_fw->g.frame.width > arg->screen_g.x
 				+ arg->screen_g.width)
 		{
@@ -990,6 +1002,7 @@ static pl_penalty_t __pl_minoverlap_get_pos_penalty(
 {
 	FvwmWindow *other_fw;
 	pl_penalty_t penalty;
+	size_borders b;
 
 	penalty = 0;
 	for (
@@ -997,7 +1010,7 @@ static pl_penalty_t __pl_minoverlap_get_pos_penalty(
 		other_fw = other_fw->next)
 	{
 		rectangle other_g;
-		Bool rc;
+		get_window_borders(other_fw, &b);
 
 		if (
 			arg->place_fw == other_fw ||
@@ -1012,7 +1025,7 @@ static pl_penalty_t __pl_minoverlap_get_pos_penalty(
 		{
 			continue;
 		}
-		rc = get_visible_window_or_icon_geometry(other_fw, &other_g);
+		(void)get_visible_window_or_icon_geometry(other_fw, &other_g);
 		if (IS_STICKY_ACROSS_PAGES(other_fw))
 		{
 			other_g.x -= arg->pdelta_p.x;
@@ -1033,6 +1046,27 @@ static pl_penalty_t __pl_minoverlap_get_pos_penalty(
 				penalty > ret->best_penalty &&
 				ret->best_penalty != -1)
 			{
+				/* TA:  20091230:  Fix over-zealous penalties
+				 * by explicitly forcing the window on-screen
+				 * here.  The y-axis is only affected here,
+				 * due to how the xoffset calculations happen
+				 * prior to setting the x-axis.  When we get
+				 * penalties which are "over-zealous" -- and
+				 * by not taking into account the size of the
+				 * window borders, the window was being placed
+				 * off screen.
+				 */
+				if (ret->best_p.y + arg->place_g.height > arg->page_p2.y)
+				{
+					ret->best_p.y =
+						(arg->page_p2.y - 
+						arg->place_g.height -
+						b.total_size.height);
+
+					ret->best_penalty = 0;
+					penalty = 0;
+				}
+
 				/* stop looking; the penalty is too high */
 				return penalty;
 			}
@@ -1206,7 +1240,7 @@ static void __place_get_placement_flags(
 		}
 		else if (reason->pos.reason != PR_POS_USE_PPOS)
 		{
-			reason->pos.reason = PR_POS_USE_USPOS;
+			reason->pos.reason = PR_POS_IGNORE_USPOS;
 		}
 	}
 	if (mode == PLACE_AGAIN)
@@ -1652,8 +1686,13 @@ static int __place_window(
 	}
 
 	if (SPLACEMENT_MODE(&pstyle->flags) != PLACE_MINOVERLAPPERCENT &&
-	    SPLACEMENT_MODE(&pstyle->flags) != PLACE_MINOVERLAP)
+	    SPLACEMENT_MODE(&pstyle->flags) != PLACE_MINOVERLAP &&
+	    SPLACEMENT_MODE(&pstyle->flags) != PLACE_POSITION)
 	{
+		/* TA:  In the case of PositionPlacement, the "ewmhiwa" option
+		 * will have already modified this for us -- so don't do it
+		 * for this placement policy.
+		 */
 		EWMH_GetWorkAreaIntersection(
 			fw, &screen_g.x, &screen_g.y, &screen_g.width,
 			&screen_g.height,
