@@ -3,11 +3,8 @@
 #ifndef _STYLE_
 #define _STYLE_
 
-#define SNAP_NONE    0x00
-#define SNAP_WINDOWS 0x01
-#define SNAP_ICONS   0x02
-#define SNAP_SAME    0x04
-#define SNAP_SCREEN  0x08
+/* The default title in case the user doesn't supply one. */
+#define DEFAULT_TITLE_FORMAT "%n"
 
 /* access to the special flags of a style */
 /* call these with a pointer to a style_flags struct */
@@ -574,13 +571,13 @@
 #define SSET_WINDOW_SHADE_STEPS(s,x) \
 	((s).shade_anim_steps = (x))
 #define SGET_SNAP_PROXIMITY(s) \
-	((s).snap_proximity)
+	((s).snap_attraction.proximity)
 #define SSET_SNAP_PROXIMITY(s,x) \
-	((s).snap_proximity = (x))
+	((s).snap_attraction.proximity = (x))
 #define SGET_SNAP_MODE(s) \
-	((s).snap_mode)
+	((s).snap_attraction.mode)
 #define SSET_SNAP_MODE(s,x) \
-	((s).snap_mode = (x))
+	((s).snap_attraction.mode = (x))
 #define SGET_SNAP_GRID_X(s) \
 	((s).snap_grid_x)
 #define SSET_SNAP_GRID_X(s,x) \
@@ -661,6 +658,14 @@
 	((s).initial_map_command_string)
 #define SSET_INITIAL_MAP_COMMAND_STRING(s,x)	\
 	((s).initial_map_command_string = (x))
+#define SGET_TITLE_FORMAT_STRING(s) \
+	((s).title_format_string)
+#define SSET_TITLE_FORMAT_STRING(s,x) \
+	((s).title_format_string = (x))
+#define SGET_ICON_TITLE_FORMAT_STRING(s) \
+	((s).icon_title_format_string)
+#define SSET_ICON_TITLE_FORMAT_STRING(s,x) \
+	((s).icon_title_format_string = (x))
 
 /* function prototypes */
 void lookup_style(FvwmWindow *fw, window_style *styles);
