@@ -22,6 +22,7 @@
 
 #include "config.h"
 #include <stdio.h>
+#include <limits.h>
 
 #include "libs/Module.h"
 #include "libs/fvwmlib.h"
@@ -46,14 +47,14 @@
 
 /* ---------------------------- local definitions --------------------------- */
 
-/* defaults for things we put in a configuration file */
-
 #define PROXY_COMMAND_DEBUG	False
 #define PROXY_GROUP_DEBUG	False
+#define PROXY_EVENT_DEBUG	False
 
 #define PROXY_KEY_POLLING	True
-#define PROXY_MOVE		False	/* move window when proxy is dragged */
-#define PROXY_ICONIFIED		False	/* show proxy when iconified */
+
+/* defaults for things we put in a configuration file */
+
 #define PROXY_WIDTH		180
 #define PROXY_HEIGHT		60
 #define PROXY_SEPARATION	10
@@ -129,8 +130,8 @@ static FlocaleFont *Ffont;
 static FlocaleFont *Ffont_small;
 static int enterSelect=False;
 static int showMiniIcons=True;
-static int proxyIconified=PROXY_ICONIFIED;
-static int proxyMove=PROXY_MOVE;
+static int proxyIconified=False;
+static int proxyMove=False;
 static int proxyWidth=PROXY_WIDTH;
 static int proxyHeight=PROXY_HEIGHT;
 static int proxySeparation=PROXY_SEPARATION;
@@ -164,6 +165,30 @@ typedef enum
 	NUMBER_OF_EXTENDED_MOUSE_BUTTONS
 } proxy_action_t;
 
+typedef enum
+{
+	PROXY_FOLLOW_OFF = 0,
+	PROXY_FOLLOW_ON,
+	PROXY_FOLLOW_INHERIT
+} proxy_follow_t;
+
+typedef enum
+{
+	PROXY_PROVOKE_RAISE = 1,
+	PROXY_PROVOKE_DESK,
+	PROXY_PROVOKE_DRAG,
+	PROXY_PROVOKE_ICON
+} proxy_provoke_t;
+
+typedef enum
+{
+	PROXY_SHOWONLY_ALL = 0,
+	PROXY_SHOWONLY_SELECT,
+	PROXY_SHOWONLY_COVER,
+	PROXY_SHOWONLY_GROUP
+} proxy_showonly_t;
+static proxy_showonly_t showOnly = PROXY_SHOWONLY_ALL;
+
 #define GROUP_COLORS 7
 static char* group_color[GROUP_COLORS][2] =
 {
@@ -245,6 +270,21 @@ static void LinkSlotAction(char *string)
 	}
 }
 
+static WindowName* FindWindowName(WindowName* namelist,char* name)
+{
+	WindowName* include=namelist;
+	while(include)
+	{
+		if(matchWildcards(include->name, name))
+		{
+			return include;
+		}
+		include=include->next;
+	}
+
+	return NULL;
+}
+
 static void LinkAction(char *string)
 {
 	char *token;
@@ -500,6 +540,28 @@ static Bool parse_options(void)
 			if (sscanf(tline, "%d", &stampLimit) < 1)
 				stampLimit=PROXY_STAMP_LIMIT;
 		}
+		else if (StrEquals(resource, "ShowOnly"))
+		{
+#if PROXY_GROUP_DEBUG
+			fprintf(stderr, "ShowOnly: \"%s\"\n", tline);
+#endif
+			if(StrEquals(tline, "All"))
+			{
+				showOnly = PROXY_SHOWONLY_ALL;
+			}
+			else if(StrEquals(tline, "Selected"))
+			{
+				showOnly = PROXY_SHOWONLY_SELECT;
+			}
+			else if(StrEquals(tline, "Covered"))
+			{
+				showOnly = PROXY_SHOWONLY_COVER;
+			}
+			else if(StrEquals(tline, "Grouped"))
+			{
+				showOnly = PROXY_SHOWONLY_GROUP;
+			}
+		}
 		else if (StrEquals(resource, "SlotWidth"))
 		{
 			if (sscanf(tline, "%d", &slotWidth) < 1)
@@ -550,23 +612,24 @@ static Bool parse_options(void)
 		}
 		else if (StrEquals(resource, "Group"))
 		{
+			ProxyGroup* proxygroup;
 			char groupname[128];
 			char directive[128];
 			char tail[128];
 			char pattern[128];
 			int bytes=0;
-			int args = sscanf(tline, "\"%[^\"]\"%s%n",
+			(void)sscanf(tline, "\"%[^\"]\"%s%n",
 				groupname,directive,&bytes);
-			ProxyGroup* proxygroup;
 
 			strncpy(tail,&tline[bytes],128);
 			tail[127]=0;
-			args = sscanf(tail, "%*[^\"]\"%[^\"]\"",pattern);
+			pattern[0]=0;
+			(void)sscanf(tail, "%*[^\"]\"%[^\"]\"",pattern);
 
 #if PROXY_GROUP_DEBUG
 			fprintf(stderr,
-				"Group: %d \"%s\" -> \"%s\" \"%s\" \"%s\"\n",
-				args,tline,groupname,directive,pattern);
+				"Group: \"%s\" -> \"%s\" \"%s\" \"%s\"\n",
+				tline,groupname,directive,pattern);
 #endif
 
 			proxygroup=FindProxyGroup(groupname);
@@ -588,12 +651,22 @@ static Bool parse_options(void)
 				proxygroup->flags.isolated=1;
 			}
 			else if (StrEquals(directive, "Include") ||
-				StrEquals(directive, "SoftInclude"))
+				StrEquals(directive, "WeakInclude") ||
+				StrEquals(directive, "SoftInclude") ||
+				StrEquals(directive, "WeakSoftInclude"))
 			{
 				WindowName* include=new_WindowName();
 				include->name=strdup(pattern);
 				include->flags.is_soft=
-					StrEquals(directive, "SoftInclude");
+					(StrEquals(directive,
+					"SoftInclude") ||
+					StrEquals(directive,
+					"WeakSoftInclude"));
+				include->flags.is_weak=
+					(StrEquals(directive,
+					"WeakInclude") ||
+					StrEquals(directive,
+					"WeakSoftInclude"));
 				include->next=proxygroup->includes;
 				proxygroup->includes=include;
 #if PROXY_GROUP_DEBUG
@@ -612,6 +685,174 @@ static Bool parse_options(void)
 					pattern);
 #endif
 			}
+			else
+			{
+/*
+			No Soft/Hard Raise/Desk/Drag
+*/
+				char* remainder = directive;
+				int soft;
+				int hard;
+				int raise;
+				int desk;
+				int drag;
+				int icon;
+
+				proxy_follow_t setting = PROXY_FOLLOW_ON;
+				if(StrHasPrefix(remainder,"No"))
+				{
+					setting = PROXY_FOLLOW_OFF;
+					remainder += 2;
+				}
+				if(StrHasPrefix(remainder,"Inherit"))
+				{
+					setting = PROXY_FOLLOW_INHERIT;
+					remainder += 7;
+				}
+
+				soft = True;
+				hard = True;
+				if(StrHasPrefix(remainder,"Soft"))
+				{
+					hard = False;
+					remainder += 4;
+				}
+				if(StrHasPrefix(remainder,"Hard"))
+				{
+					soft = False;
+					remainder += 4;
+				}
+
+				raise = True;
+				desk = True;
+				drag = True;
+				icon = True;
+				if(StrHasPrefix(remainder,"Raise"))
+				{
+					desk = False;
+					drag = False;
+					icon = False;
+					remainder += 5;
+				}
+				if(StrHasPrefix(remainder,"Desk"))
+				{
+					raise = False;
+					drag = False;
+					icon = False;
+					remainder += 4;
+				}
+				if(StrHasPrefix(remainder,"Drag"))
+				{
+					raise = False;
+					desk = False;
+					icon = False;
+					remainder += 4;
+				}
+				if(StrHasPrefix(remainder,"Icon"))
+				{
+					raise = False;
+					desk = False;
+					drag = False;
+					remainder += 4;
+				}
+				if(StrHasPrefix(remainder,"All"))
+				{
+					remainder += 3;
+				}
+#if PROXY_GROUP_DEBUG
+				fprintf(stderr,"@@@ ");
+				fprintf(stderr,"Follow \"%s\"\n",
+					pattern);
+				fprintf(stderr,"@@@ ");
+				fprintf(stderr,"  setting %d soft %d hard %d\n",
+					setting, soft, hard);
+				fprintf(stderr,"@@@ ");
+				fprintf(stderr,"  raise %d desk %d",
+					raise, desk);
+				fprintf(stderr," drag %d icon %d\n",
+					drag, icon);
+#endif
+
+				if(pattern[0])
+				{
+					WindowName* windowName = FindWindowName(
+						proxygroup->includes, pattern);
+					if(windowName)
+					{
+						if(soft)
+						{
+							if(raise) windowName->
+								flags.
+								soft_raise =
+								setting;
+							if(desk) windowName->
+								flags.
+								soft_desk =
+								setting;
+							if(drag) windowName->
+								flags.
+								soft_drag =
+								setting;
+							if(icon) windowName->
+								flags.
+								soft_icon =
+								setting;
+						}
+						if(hard)
+						{
+							if(raise) windowName->
+								flags.
+								hard_raise =
+								setting;
+							if(desk) windowName->
+								flags.
+								hard_desk =
+								setting;
+							if(drag) windowName->
+								flags.
+								hard_drag =
+								setting;
+							if(icon) windowName->
+								flags.
+								hard_icon =
+								setting;
+						}
+					}
+				}
+				else
+				{
+					if(soft)
+					{
+						if(raise) proxygroup->
+							flags.soft_raise =
+							setting;
+						if(desk) proxygroup->
+							flags.soft_desk =
+							setting;
+						if(drag) proxygroup->
+							flags.soft_drag =
+							setting;
+						if(icon) proxygroup->
+							flags.soft_icon =
+							setting;
+					}
+					if(hard)
+					{
+						if(raise) proxygroup->
+							flags.hard_raise =
+							setting;
+						if(desk) proxygroup->
+							flags.hard_desk =
+							setting;
+						if(drag) proxygroup->
+							flags.hard_drag =
+							setting;
+						if(icon) proxygroup->
+							flags.hard_icon =
+							setting;
+					}
+				}
+			}
 		}
 		else
 		{
@@ -663,6 +904,15 @@ static void delete_ProxyWindow(ProxyWindow *p)
 static void WindowName_WindowName(WindowName *p)
 {
 	memset(p, 0, sizeof *p);
+
+	p->flags.soft_raise = PROXY_FOLLOW_INHERIT;
+	p->flags.soft_desk = PROXY_FOLLOW_INHERIT;
+	p->flags.soft_drag = PROXY_FOLLOW_INHERIT;
+	p->flags.soft_icon = PROXY_FOLLOW_INHERIT;
+	p->flags.hard_raise = PROXY_FOLLOW_INHERIT;
+	p->flags.hard_desk = PROXY_FOLLOW_INHERIT;
+	p->flags.hard_drag = PROXY_FOLLOW_INHERIT;
+	p->flags.hard_icon = PROXY_FOLLOW_INHERIT;
 }
 
 static WindowName *new_WindowName(void)
@@ -675,6 +925,15 @@ static WindowName *new_WindowName(void)
 static void ProxyGroup_ProxyGroup(ProxyGroup *p)
 {
 	memset(p, 0, sizeof *p);
+
+	p->flags.soft_raise = PROXY_FOLLOW_ON;
+	p->flags.soft_desk = PROXY_FOLLOW_ON;
+	p->flags.soft_drag = PROXY_FOLLOW_ON;
+	p->flags.soft_icon = PROXY_FOLLOW_ON;
+	p->flags.hard_raise = PROXY_FOLLOW_ON;
+	p->flags.hard_desk = PROXY_FOLLOW_ON;
+	p->flags.hard_drag = PROXY_FOLLOW_ON;
+	p->flags.hard_icon = PROXY_FOLLOW_ON;
 }
 
 static ProxyGroup *new_ProxyGroup(void)
@@ -735,7 +994,6 @@ static void send_command_to_fvwm(char *command, Window w)
 static int GetProperty(Window w,char* propertyname)
 {
 	Atom atom,actual_type;
-	char *atom_name;
 	int actual_format;
 	unsigned long nitems;
 	unsigned long bytes_after;
@@ -746,7 +1004,6 @@ static int GetProperty(Window w,char* propertyname)
 	int bytes;
 
 	atom = XInternAtom(dpy, propertyname, True);
-	atom_name = XGetAtomName (dpy,atom);
 
 	status = XGetWindowProperty(dpy, w, atom, 0L, 1024,
 		False, AnyPropertyType,
@@ -798,7 +1055,6 @@ static int GetLeader(Window w)
 static int GetParentProcessId(int pid)
 {
 	int ppid=0;
-	int bytes;
 	FILE* statusfile;
 
 	sprintf(commandBuffer,"/proc/%d/stat",pid);
@@ -807,7 +1063,7 @@ static int GetParentProcessId(int pid)
 	{
 		return 0;
 	}
-	bytes=fread(resultBuffer,32,1,statusfile);
+	(void)fread(resultBuffer,32,1,statusfile);
 	sscanf(resultBuffer,"%*d %*[^)]) %*s %d",&ppid);
 	fclose(statusfile);
 	return ppid;
@@ -833,17 +1089,8 @@ static ProxyGroup* FindProxyGroup(char* groupname)
 
 static int MatchWindowName(WindowName* namelist,char* name)
 {
-	WindowName* include=namelist;
-	while(include)
-	{
-		if(matchWildcards(include->name, name))
-		{
-			return 1;
-		}
-		include=include->next;
-	}
-
-	return 0;
+	WindowName* include=FindWindowName(namelist, name);
+	return (include != NULL);
 }
 
 static ProxyGroup* FindProxyGroupWithWindowName(char* name)
@@ -868,6 +1115,30 @@ static ProxyGroup* FindProxyGroupWithWindowName(char* name)
 	return NULL;
 }
 
+static ProxyGroup* FindProxyGroupOfNeighbor(ProxyWindow* instigator)
+{
+	ProxyWindow *proxy;
+
+	if(!instigator->group)
+	{
+		return NULL;
+	}
+
+	for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+	{
+		if(proxy==instigator)
+		{
+			continue;
+		}
+		if (proxy->desk == deskNumber &&
+			proxy->group==instigator->group && proxy->proxy_group)
+		{
+			return proxy->proxy_group;
+		}
+	}
+	return NULL;
+}
+
 static ProxyWindow *FindProxy(Window window)
 {
 	ProxyWindow *proxy=firstProxy;
@@ -952,6 +1223,15 @@ static void DrawWindow(
 	}
 	RelieveRectangle(dpy,proxy->proxy, 0,0, w - 1,h - 1, hi_gc,sh_gc, 2);
 
+#if 0
+	if (proxy->iconname != NULL)
+	{
+		free(proxy->iconname);
+	}
+	proxy->iconname = safestrdup("    ");
+	sprintf(proxy->iconname, "%d", proxy->stack);
+#endif
+
 	big_name = proxy->iconname;
 	if(big_name==NULL || !big_name[0])
 	{
@@ -1105,18 +1385,48 @@ static void OpenOneWindow(ProxyWindow *proxy)
 	unsigned long valuemask=CWOverrideRedirect;
 	XSetWindowAttributes attributes;
 
-	if (proxy == NULL)
+	if (proxy == NULL ||
+		proxy->desk != deskNumber ||
+		(!proxyIconified && proxy->flags.is_iconified) ||
+		proxy->flags.is_shown)
 	{
 		return;
 	}
-	if (proxy->desk != deskNumber ||
-			(!proxyIconified && proxy->flags.is_iconified) )
+	if (showOnly)
 	{
-		return;
-	}
-	if (proxy->flags.is_shown)
-	{
-		return;
+		if (!focusWindow)
+		{
+			return;
+		}
+		if ((!selectProxy || proxy != selectProxy) &&
+			proxy->window != focusWindow)
+		{
+			ProxyWindow *other;
+			if (showOnly == PROXY_SHOWONLY_SELECT)
+			{
+				return;
+			}
+			other = FindProxy(focusWindow);
+#if PROXY_GROUP_DEBUG
+			fprintf(stderr, "OpenOneWindow"
+				" %d %d %d %d  %d %d %d %d  %d %d\n",
+				proxy->proxyx, proxy->proxyy,
+				proxy->proxyw, proxy->proxyh,
+				other->x, other->y,
+				other->w, other->h,
+				proxy->group, other->group);
+#endif
+			if ((showOnly == PROXY_SHOWONLY_COVER ||
+				!other->group ||
+				proxy->group != other->group) &&
+				(proxy->proxyx > other->x+other->w ||
+				proxy->proxyx+proxy->proxyw < other->x ||
+				proxy->proxyy > other->y+other->h ||
+				proxy->proxyy+proxy->proxyh < other->y))
+			{
+				return;
+			}
+		}
 	}
 	if (proxy->proxy == None)
 	{
@@ -1263,6 +1573,69 @@ static Bool AdjustOneWindow(ProxyWindow *proxy)
 				dy<proxyHeight+proxySeparation )
 		{
 			rc = True;
+#if PROXY_GROUP_DEBUG
+			fprintf(stderr,
+				"AdjustOneWindow %d %d %d %d  %d %d %d %d\n",
+				proxy->proxyx, proxy->proxyy,
+				proxy->proxyw, proxy->proxyh,
+				other->proxyx, other->proxyy,
+				other->proxyw, other->proxyh);
+#endif
+			if (proxyWidth-dx<proxyHeight-dy)
+			{
+				if (proxy->proxyx<=other->proxyx)
+				{
+					int delta = proxy->proxyx +
+						proxy->proxyw +
+						proxySeparation -
+						other->proxyx;
+					proxy->proxyx -= delta/2;
+					other->proxyx += delta - delta/2;
+#if PROXY_GROUP_DEBUG
+					fprintf(stderr, "left %d\n", delta);
+#endif
+				}
+				else
+				{
+					int delta = other->proxyx +
+						other->proxyw +
+						proxySeparation -
+						proxy->proxyx;
+					other->proxyx -= delta/2;
+					proxy->proxyx += delta - delta/2;
+#if PROXY_GROUP_DEBUG
+					fprintf(stderr, "right %d\n", delta);
+#endif
+				}
+			}
+			else
+			{
+				if (proxy->proxyy<=other->proxyy)
+				{
+					int delta = proxy->proxyy +
+						proxy->proxyh +
+						proxySeparation -
+						other->proxyy;
+					proxy->proxyy -= delta/2;
+					other->proxyy += delta - delta/2;
+#if PROXY_GROUP_DEBUG
+					fprintf(stderr, "up %d\n", delta);
+#endif
+				}
+				else
+				{
+					int delta = other->proxyy +
+						other->proxyh +
+						proxySeparation -
+						proxy->proxyy;
+					other->proxyy -= delta/2;
+					proxy->proxyy += delta - delta/2;
+#if PROXY_GROUP_DEBUG
+					fprintf(stderr, "down %d\n", delta);
+#endif
+				}
+			}
+/*
 			if (proxyWidth-dx<proxyHeight-dy)
 			{
 				if (proxy->proxyx<=other->proxyx)
@@ -1293,6 +1666,7 @@ static Bool AdjustOneWindow(ProxyWindow *proxy)
 						proxySeparation;
 				}
 			}
+*/
 		}
 	}
 
@@ -1401,6 +1775,104 @@ static void WaitToConfig(ProxyWindow *proxy)
 	}
 }
 
+static int Provokable(ProxyWindow *proxy,proxy_provoke_t provocation)
+{
+	int soft=proxy->flags.is_soft;
+	ProxyGroup* proxy_group=proxy->proxy_group;
+	proxy_follow_t follow;
+	WindowName* include;
+#if PROXY_GROUP_DEBUG
+	fprintf(stderr, "Provokable %p %s %x soft %d\n",
+		proxy,proxy->name,provocation,soft);
+#endif
+	if(!proxy_group)
+	{
+		proxy_group=FindProxyGroupOfNeighbor(proxy);
+	}
+	if(!proxy_group)
+	{
+		return True;
+	}
+
+	follow=PROXY_FOLLOW_INHERIT;
+
+#if PROXY_GROUP_DEBUG
+	fprintf(stderr, "  group flags %x\n",proxy_group->flags);
+#endif
+	include = FindWindowName(proxy_group->includes,proxy->name);
+	if(include)
+	{
+#if PROXY_GROUP_DEBUG
+		fprintf(stderr, "  include flags %x\n",include->flags);
+#endif
+
+		switch(provocation)
+		{
+			case PROXY_PROVOKE_RAISE:
+				follow=soft? include->flags.soft_raise:
+					include->flags.hard_raise;
+			break;
+
+			case PROXY_PROVOKE_DESK:
+				follow=soft? include->flags.soft_desk:
+					include->flags.hard_desk;
+			break;
+
+			case PROXY_PROVOKE_DRAG:
+				follow=soft? include->flags.soft_drag:
+					include->flags.hard_drag;
+			break;
+
+			case PROXY_PROVOKE_ICON:
+				follow=soft? include->flags.soft_icon:
+					include->flags.hard_icon;
+			break;
+		}
+#if PROXY_GROUP_DEBUG
+		fprintf(stderr, "  include follow %d\n",follow);
+#endif
+		if(follow!=PROXY_FOLLOW_INHERIT)
+		{
+			return (follow!=PROXY_FOLLOW_OFF);
+		}
+	}
+
+	follow=PROXY_FOLLOW_INHERIT;
+	switch(provocation)
+	{
+		case PROXY_PROVOKE_RAISE:
+			follow=soft? proxy_group->flags.soft_raise:
+				proxy_group->flags.hard_raise;
+		break;
+
+		case PROXY_PROVOKE_DESK:
+			follow=soft? proxy_group->flags.soft_desk:
+				proxy_group->flags.hard_desk;
+		break;
+
+		case PROXY_PROVOKE_DRAG:
+			follow=soft? proxy_group->flags.soft_drag:
+				proxy_group->flags.hard_drag;
+		break;
+
+		case PROXY_PROVOKE_ICON:
+			follow=soft? proxy_group->flags.soft_icon:
+				proxy_group->flags.hard_icon;
+		break;
+	}
+#if PROXY_GROUP_DEBUG
+	fprintf(stderr, "  group follow %d\n",follow);
+#endif
+	return (follow!=PROXY_FOLLOW_OFF);
+}
+
+static int Provoked(ProxyWindow *instigator,ProxyWindow *proxy,
+	proxy_provoke_t provocation)
+{
+	return (Provokable(instigator,provocation) &&
+		Provokable(proxy,provocation));
+}
+
 static void IsolateGroup(ProxyWindow *instigator,int isolate)
 {
 	ProxyWindow *proxy;
@@ -1452,6 +1924,10 @@ static void IconifyGroup(ProxyWindow *instigator,int iconify)
 		{
 			continue;
 		}
+		if(!Provoked(instigator,proxy,PROXY_PROVOKE_ICON))
+		{
+			continue;
+		}
 		if (proxy->desk == deskNumber &&
 			proxy->group==instigator->group)
 		{
@@ -1462,12 +1938,316 @@ static void IconifyGroup(ProxyWindow *instigator,int iconify)
 	}
 }
 
+void RefineStack(int desired)
+{
+	ProxyWindow *proxy;
+	ProxyWindow *best = NULL;
+	int index = 0;
+#if PROXY_GROUP_DEBUG
+	fprintf(stderr, "RefineStack %d\n", desired);
+#endif
+	for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+	{
+		proxy->stack_tmp = -1;
+	}
+	while(True)
+	{
+		best = NULL;
+		for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+		{
+			if(proxy->stack_tmp < 0 && (best == NULL ||
+				(desired? proxy->stack_desired: proxy->stack) <
+				(desired? best->stack_desired: best->stack)))
+			{
+				best = proxy;
+			}
+		}
+		if(best == NULL)
+		{
+			break;
+		}
+#if PROXY_GROUP_DEBUG
+		fprintf(stderr, "%p %d %d -> %d %s\n", best,
+			best->stack, best->stack_desired,
+			index, best->name);
+#endif
+		best->stack_tmp = index++;
+	}
+	for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+	{
+		if(desired)
+		{
+			proxy->stack_desired = proxy->stack_tmp;
+		}
+		else
+		{
+			proxy->stack = proxy->stack_tmp;
+		}
+	}
+}
+
+/* doesn't work: non-fatal X error in X_ConfigureWindow */
+void RestackAtomic(int raise)
+{
+	ProxyWindow *proxy;
+	int count = 0;
+	Window *windows;
+	RefineStack(False);
+
+	for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+	{
+		count++;
+	}
+	windows = (Window *) safemalloc (count * sizeof (Window));
+	for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+	{
+		fprintf(stderr, "RestackAtomic %d %s\n",
+			proxy->stack, proxy->name);
+		windows[proxy->stack] = proxy->window;
+	}
+	XRestackWindows(dpy, windows, count);
+	free (windows);
+}
+
+void RestackIncremental(int raise)
+{
+	int changes = 0;
+	int changed = 1;
+
+	RefineStack(False);
+	RefineStack(True);
+
+	while(changed)
+	{
+		ProxyWindow *proxy;
+		ProxyWindow *best = NULL;
+		int next_delta = 0;
+		changed = 0;
+
+#if PROXY_GROUP_DEBUG
+		for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+		{
+			fprintf(stderr, "%2d ", proxy->stack);
+		}
+		fprintf(stderr, "\n");
+		for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+		{
+			fprintf(stderr, "%2d ", proxy->stack_desired);
+		}
+		fprintf(stderr, "\n");
+#endif
+		for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+		{
+			int delta = raise?
+				proxy->stack - proxy->stack_desired:
+				proxy->stack_desired - proxy->stack;
+			if(delta > 0 &&
+				(best == NULL || delta > next_delta))
+			{
+				best = proxy;
+				next_delta = delta;
+			}
+		}
+		if(best)
+		{
+#if PROXY_GROUP_DEBUG
+			fprintf(stderr,
+				"RestackIncremental Raise %p %s %d -> %d\n",
+				best, best->name,
+				best->stack, best->stack_desired);
+#endif
+			if(raise)
+			{
+				best->raised=1;
+				XRaiseWindow(dpy, best->window);
+				best->stack = INT_MIN;
+			}
+			else
+			{
+				best->raised= -1;
+				XLowerWindow(dpy, best->window);
+				best->stack = INT_MAX;
+			}
+			RefineStack(False);
+			changed = 1;
+		}
+		if(++changes > 32)
+		{
+#if PROXY_GROUP_DEBUG
+			fprintf(stderr,
+				"RestackIncremental excessive changes (%d)\n",
+				changes);
+#endif
+			break;
+		}
+#if PROXY_GROUP_DEBUG
+		fprintf(stderr, "*****************************************\n");
+		fprintf(stderr, "RestackIncremental(%d) %d changes\n",
+			raise, changes);
+#endif
+	}
+}
+
+void Restack(int raise)
+{
+	RestackIncremental(raise);
+}
+
+void ReadStack(unsigned long *body, unsigned long length)
+{
+	Window window;
+	ProxyWindow *proxy;
+	int i;
+	ProxyWindow *instigator=NULL;
+	int index;
+
+#if PROXY_GROUP_DEBUG
+	fprintf(stderr, "ReadStack %p %d-%d\n",
+		body, (int)length, FvwmPacketHeaderSize);
+#endif
+
+	for (i = 0; i < (length - FvwmPacketHeaderSize); i += 3)
+	{
+		instigator=FindProxy(body[0]);
+		if(instigator)
+		{
+			break;
+		}
+	}
+	if(!instigator)
+	{
+		return;
+	}
+	for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+	{
+		if(proxy->stack>instigator->stack)
+		{
+			proxy->stack+=10000;
+		}
+	}
+
+	index=instigator->stack;
+	for (i = 0; i < (length - FvwmPacketHeaderSize); i += 3)
+	{
+		window = body[i];
+#if PROXY_GROUP_DEBUG
+		fprintf(stderr, "  stack %d %d\n", i, (int)window);
+#endif
+		proxy = FindProxy(window);
+		if(proxy)
+		{
+#if PROXY_GROUP_DEBUG
+			fprintf(stderr, " %p %s",proxy,proxy->name);
+#endif
+			proxy->stack = index++;
+		}
+#if PROXY_GROUP_DEBUG
+		fprintf(stderr, "\n");
+#endif
+	}
+	RefineStack(False);
+}
+
 static void RaiseLowerGroup(Window w,int raise)
 {
 	ProxyWindow *instigator;
 	ProxyWindow *proxy;
 
 	instigator = FindProxy(w);
+#if PROXY_GROUP_DEBUG
+	fprintf(stderr, "RaiseLowerGroup %d %s\n", raise,
+			instigator? instigator->name: "<none>");
+#endif
+	if(instigator==NULL)
+	{
+		return;
+	}
+	instigator->stack = raise? INT_MIN: INT_MAX;
+	RefineStack(False);
+	if(instigator->flags.is_isolated)
+	{
+		return;
+	}
+	if(!Provokable(instigator,PROXY_PROVOKE_RAISE))
+	{
+		return;
+	}
+
+#if PROXY_GROUP_DEBUG
+	fprintf(stderr, "RaiseLowerGroup %p %d %d\n",
+		instigator,raise,instigator->raised);
+#endif
+
+	if(abs(instigator->raised)>10)
+	{
+		exit(1);
+	}
+
+	if(instigator->raised == (raise? 1: -1))
+	{
+		return;
+	}
+
+	instigator->raised=0;
+
+	if(instigator->flags.is_iconified || !instigator->group)
+	{
+		return;
+	}
+
+	instigator->stack = raise? INT_MIN: INT_MAX;
+	for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+	{
+		proxy->stack_desired = proxy->stack;
+	}
+	for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
+	{
+		if(proxy==instigator)
+		{
+			continue;
+		}
+		if(!Provokable(proxy,PROXY_PROVOKE_RAISE))
+		{
+			continue;
+		}
+		if (proxy->desk == deskNumber &&
+			proxy->group==instigator->group)
+		{
+			if(raise)
+			{
+#if PROXY_GROUP_DEBUG
+				fprintf(stderr, "Raise %p\n",proxy);
+#endif
+
+				proxy->stack_desired-=10000;
+			}
+			else
+			{
+#if PROXY_GROUP_DEBUG
+				fprintf(stderr, "Lower %p\n",proxy);
+#endif
+
+				proxy->stack_desired+=10000;
+			}
+		}
+	}
+	Restack(raise);
+}
+
+/*
+static void RaiseLowerGroupOld(Window w,int raise)
+{
+	ProxyWindow *instigator;
+	ProxyWindow *proxy;
+
+	instigator = FindProxy(w);
+	fprintf(stderr, "RaiseLowerGroup %d %s\n", raise,
+			instigator? instigator->name: "<none>");
+	if(instigator)
+	{
+		instigator->stack = raise? INT_MIN: INT_MAX;
+		RefineStack(False);
+	}
 	if(instigator==NULL || instigator->flags.is_isolated)
 	{
 		return;
@@ -1525,7 +2305,6 @@ static void RaiseLowerGroup(Window w,int raise)
 			}
 		}
 	}
-	/* make sure instigator ends on on top/bottom */
 	if(raise)
 	{
 #if PROXY_GROUP_DEBUG
@@ -1545,6 +2324,7 @@ static void RaiseLowerGroup(Window w,int raise)
 		XLowerWindow(dpy, instigator->window);
 	}
 }
+*/
 
 static void ClearRaised(void)
 {
@@ -1581,6 +2361,11 @@ static void ShiftWindows(ProxyWindow *instigator,int dx,int dy)
 		{
 			continue;
 		}
+		if(!instigator->flags.is_isolated &&
+			!Provoked(instigator,proxy,PROXY_PROVOKE_DRAG))
+		{
+			continue;
+		}
 		if (proxy->desk == deskNumber &&
 			proxy->group==instigator->group)
 		{
@@ -1623,6 +2408,10 @@ static void CatchWindows(ProxyWindow *instigator,int vertical,int from,int to,
 		{
 			continue;
 		}
+		if(!Provoked(instigator,proxy,PROXY_PROVOKE_DRAG))
+		{
+			continue;
+		}
 		if (proxy->desk == deskNumber &&
 			proxy->group==instigator->group &&
 			!proxy->flags.is_soft)
@@ -1837,6 +2626,10 @@ static void MoveGroupToDesk(ProxyWindow *instigator,int desk)
 	sprintf(commandBuffer, "MoveToDesk 0 %d", desk);
 	for (proxy=firstProxy; proxy != NULL; proxy=proxy->next)
 	{
+		if(!Provoked(instigator,proxy,PROXY_PROVOKE_DESK))
+		{
+			continue;
+		}
 		if (proxy->desk == old_desk &&
 			proxy->group==old_group)
 		{
@@ -1964,6 +2757,11 @@ static ProxyWindow* FindNeighborForProcess(ProxyWindow* proxy,
 	int group;
 	int auto_include=0;
 
+#if PROXY_GROUP_DEBUG
+		fprintf(stderr,"FindNeighborForProcess %p desk %d pid %d %d\n",
+			proxy, desk, ppid, pid);
+#endif
+
 	if(!pid)
 	{
 		return 0;
@@ -1979,8 +2777,8 @@ static ProxyWindow* FindNeighborForProcess(ProxyWindow* proxy,
 				proxy->name) ||
 			(auto_include=other->proxy_group->flags.auto_include)))
 		{
-			proxy->flags.is_soft=auto_include &&
-				other->proxy_group->flags.auto_soft;
+			proxy->flags.is_soft = (auto_include &&
+				other->proxy_group->flags.auto_soft);
 			if(other->proxy_group->flags.isolated &&
 				!proxy->flags.is_isolated)
 			{
@@ -2041,6 +2839,11 @@ static ProxyWindow* FindNeighborForLeader(ProxyWindow* proxy,
 	int group;
 	int auto_include=0;
 
+#if PROXY_GROUP_DEBUG
+		fprintf(stderr,"FindNeighborForLeader %p desk %d leader %d \n",
+			proxy, desk, leader);
+#endif
+
 	if(!leader)
 	{
 		return 0;
@@ -2123,16 +2926,17 @@ static ProxyWindow* FindNeighborForApplication(ProxyWindow* proxy)
 static int ProxyGroupCheckSoft(ProxyGroup* proxy_group,
 	ProxyWindow* proxy)
 {
-	WindowName* include=proxy_group->includes;
-	while(include)
-	{
-		if(matchWildcards(include->name, proxy->name))
-		{
-			return include->flags.is_soft;
-		}
-		include=include->next;
-	}
-	return 0;
+	WindowName* include=FindWindowName(proxy_group->includes,
+		proxy->name);
+	return include? include->flags.is_soft: 0;
+}
+
+static int ProxyGroupCheckWeak(ProxyGroup* proxy_group,
+	ProxyWindow* proxy)
+{
+	WindowName* include=FindWindowName(proxy_group->includes,
+		proxy->name);
+	return include? include->flags.is_weak: 0;
 }
 
 static void UpdateProxyGroup(ProxyWindow* proxy)
@@ -2163,6 +2967,11 @@ static void UpdateProxyGroup(ProxyWindow* proxy)
 	{
 		proxy->proxy_group=
 			FindProxyGroupWithWindowName(proxy->name);
+		if(proxy->proxy_group &&
+			ProxyGroupCheckWeak(proxy->proxy_group,proxy))
+		{
+			proxy->proxy_group=NULL;
+		}
 	}
 	if(proxy->proxy_group &&
 		MatchWindowName(proxy->proxy_group->excludes,proxy->name))
@@ -2785,6 +3594,11 @@ static void MarkProxy(ProxyWindow *new_proxy)
 		send_command_to_fvwm(action_list[PROXY_ACTION_MARK],
 				selectProxy->window);
 
+	if(showOnly && are_windows_shown)
+	{
+		ReshuffleWindows();
+		selectProxy = new_proxy;
+	}
 	return;
 }
 
@@ -2924,6 +3738,7 @@ static void change_cset(int cset)
 static void ProcessMessage(FvwmPacket* packet)
 {
 	unsigned long type = packet->type;
+	unsigned long length = packet->size;
 	unsigned long* body = packet->body;
 	FvwmWinPacketBodyHeader *bh = (void *)body;
 	ProxyWindow *proxy;
@@ -2931,8 +3746,9 @@ static void ProcessMessage(FvwmPacket* packet)
 	int y=0;
 	int w=0;
 	int h=0;
+	int reshuffle=0;
 
-	if(type!=M_RAISE_WINDOW && type!=M_LOWER_WINDOW &&
+	if(type!=M_RAISE_WINDOW && type!=M_LOWER_WINDOW && type!=M_RESTACK &&
 		type!=M_FOCUS_CHANGE && type!=M_ICON_NAME &&
 		type!=M_MINI_ICON && type!=M_STRING)
 	{
@@ -2942,25 +3758,59 @@ static void ProcessMessage(FvwmPacket* packet)
 	switch (type)
 	{
 	case M_CONFIGURE_WINDOW:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_CONFIGURE_WINDOW\n");
+#endif
 	case M_ADD_WINDOW:
+#if PROXY_EVENT_DEBUG
+		if(type!=M_CONFIGURE_WINDOW)
+		{
+			fprintf(stderr,
+				"FvwmProxy ProcessMessage M_ADD_WINDOW\n");
+		}
+#endif
 		ConfigureWindow(packet);
 		break;
 	case M_DESTROY_WINDOW:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_DESTROY_WINDOW\n");
+#endif
 		DestroyWindow(bh->w);
 		break;
 	case M_ICONIFY:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_ICONIFY\n");
+#endif
 		IconifyWindow(FindProxy(bh->w), 1);
 		break;
 	case M_DEICONIFY:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_DEICONIFY\n");
+#endif
 		IconifyWindow(FindProxy(bh->w), 0);
 		break;
 	case M_RAISE_WINDOW:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_RAISE_WINDOW\n");
+#endif
 		RaiseLowerGroup(bh->w,1);
 		break;
 	case M_LOWER_WINDOW:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_LOWER_WINDOW\n");
+#endif
 		RaiseLowerGroup(bh->w,0);
 		break;
+	case M_RESTACK:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_RESTACK\n");
+#endif
+		ReadStack(body,length);
+		break;
 	case M_WINDOW_NAME:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_WINDOW_NAME\n");
+#endif
 		proxy = FindProxy(bh->w);
 		if (proxy != NULL)
 		{
@@ -2977,6 +3827,9 @@ static void ProcessMessage(FvwmPacket* packet)
 		}
 		break;
 	case M_ICON_NAME:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_ICON_NAME\n");
+#endif
 		proxy = FindProxy(bh->w);
 		if (proxy != NULL)
 		{
@@ -2995,6 +3848,9 @@ static void ProcessMessage(FvwmPacket* packet)
 		}
 		break;
 	case M_NEW_DESK:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_NEW_DESK\n");
+#endif
 		if (deskNumber!=body[0])
 		{
 			deskNumber=body[0];
@@ -3005,10 +3861,16 @@ static void ProcessMessage(FvwmPacket* packet)
 		}
 		break;
 	case M_NEW_PAGE:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_NEW_PAGE\n");
+#endif
 		deskNumber=body[2];
 		ReshuffleWindows();
 		break;
 	case M_MINI_ICON:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_MINI_ICON\n");
+#endif
 		proxy = FindProxy(bh->w);
 		if (proxy != NULL)
 		{
@@ -3022,24 +3884,33 @@ static void ProcessMessage(FvwmPacket* packet)
 		}
 		break;
 	case M_FOCUS_CHANGE:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_FOCUS_CHANGE\n");
+#endif
 	{
 		focusWindow=bh->w;
 		if(bh->w != 0)
 		{
 			last_rotation_instigator=NULL;
 		}
+		if(showOnly && !selectProxy)
+		{
+			reshuffle=1;
+		}
 	}
 		break;
 	case M_STRING:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_STRING\n");
+#endif
 	{
 		char *message=(char*)&body[3];
 		char *token;
 		char *next;
 		int prev;
 		int isolate_check=0;
-		int reshuffle=0;
 
-#if PROXY_GROUP_DEBUG
+#if PROXY_EVENT_DEBUG
 		fprintf(stderr, "M_STRING \"%s\"\n", message);
 #endif
 
@@ -3214,17 +4085,12 @@ static void ProcessMessage(FvwmPacket* packet)
 		{
 			IsolateCheck(proxy,0);
 		}
-		if(reshuffle)
-		{
-			/* windows may have moved, so update proxy windows */
-			if (are_windows_shown)
-			{
-				ReshuffleWindows();
-			}
-		}
 		break;
 	}
 	case M_CONFIG_INFO:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy ProcessMessage M_CONFIG_INFO\n");
+#endif
 	{
 		char *tline, *token;
 
@@ -3240,6 +4106,15 @@ static void ProcessMessage(FvwmPacket* packet)
 	}
 	}
 
+	if(reshuffle)
+	{
+		/* windows may have moved, so update proxy windows */
+		if (are_windows_shown)
+		{
+			ReshuffleWindows();
+		}
+	}
+
 	return;
 }
 
@@ -3296,6 +4171,9 @@ static void DispatchEvent(XEvent *pEvent)
 	switch(pEvent->xany.type)
 	{
 	case Expose:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy DispatchEvent Expose\n");
+#endif
 		proxy = FindProxy(window);
 		if (proxy != NULL)
 		{
@@ -3305,6 +4183,9 @@ static void DispatchEvent(XEvent *pEvent)
 		}
 		break;
 	case ButtonPress:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy DispatchEvent ButtonPress\n");
+#endif
 		proxy = FindProxy(window);
 		if (proxy)
 		{
@@ -3352,6 +4233,9 @@ static void DispatchEvent(XEvent *pEvent)
 		mousey=pEvent->xbutton.y_root;
 		break;
 	case MotionNotify:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy DispatchEvent MotionNotify\n");
+#endif
 		proxy = FindProxy(window);
 		dx=pEvent->xbutton.x_root-mousex;
 		dy=pEvent->xbutton.y_root-mousey;
@@ -3365,6 +4249,9 @@ static void DispatchEvent(XEvent *pEvent)
 		mousey=pEvent->xbutton.y_root;
 		break;
 	case EnterNotify:
+#if PROXY_EVENT_DEBUG
+		fprintf(stderr, "FvwmProxy DispatchEvent EnterNotify\n");
+#endif
 		proxy = FindProxy(pEvent->xcrossing.window);
 		if (pEvent->xcrossing.mode == NotifyNormal)
 		{
@@ -3469,7 +4356,7 @@ int main(int argc, char **argv)
 		M_CONFIGURE_WINDOW| M_ADD_WINDOW| M_FOCUS_CHANGE|
 		M_DESTROY_WINDOW| M_NEW_DESK| M_NEW_PAGE| M_ICON_NAME|
 		M_WINDOW_NAME| M_MINI_ICON| M_ICONIFY| M_DEICONIFY|
-		M_CONFIG_INFO| M_END_CONFIG_INFO);
+		M_CONFIG_INFO| M_END_CONFIG_INFO| M_RESTACK);
 
 	if (parse_options() == False)
 		exit(1);
